import threading
from abc import abstractmethod
from concurrent.futures import as_completed

from fsd_utils.services.aws_extended_client import SQSExtendedClient


class TaskExecutorService:
    def __init__(
        self,
        flask_app,
        executor,
        s3_bucket,
        sqs_primary_url,
        task_executor_max_thread,
        sqs_batch_size,
        visibility_time,
        sqs_wait_time,
        endpoint_url_override=None,
        aws_access_key_id=None,
        aws_secret_access_key=None,
        region_name=None,
    ):
        self.executor = executor
        self.sqs_primary_url = sqs_primary_url
        self.task_executor_max_thread = task_executor_max_thread
        self.sqs_batch_size = sqs_batch_size
        self.visibility_time = visibility_time
        self.sqs_wait_time = sqs_wait_time
        self.logger = flask_app.logger
        self.sqs_extended_client = SQSExtendedClient(
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            region_name=region_name,
            endpoint_url=endpoint_url_override,
            large_payload_support=s3_bucket,
            always_through_s3=True,
            delete_payload_from_s3=True,
            logger=self.logger,
        )
        self.logger.info("Created the thread pool executor to process messages in extended SQS queue")

    def process_messages(self):
        """
        Scheduler calling this method based on a cron job for every given second then messages will be read
        from the SQS queue in AWS and if S3 usage is allowed then it will interact each other to retrieve the messages
        """
        current_thread = threading.current_thread()
        thread_id = f"[{current_thread.name}:{current_thread.ident}]"
        self.logger.debug("%s Triggered schedular to get messages", thread_id)

        running_threads, read_msg_ids = self._handle_message_receiving_and_processing()

        self._handle_message_delete_processing(running_threads, read_msg_ids)

        self.logger.debug("%s Message Processing completed and will start again later", thread_id)

    @abstractmethod
    def message_executor(self, message):
        """
        Processing the message in a separate worker thread and this will call the GOV notify service to send emails
        :param message Json message
        override this for implementation
        """
        pass

    def _handle_message_receiving_and_processing(self):
        """
        Handle message retrieve from the SQS service and get the json from S3 bucket
        """
        current_thread = threading.current_thread()
        thread_id = f"[{current_thread.name}:{current_thread.ident}]"
        running_threads = []
        read_msg_ids = []
        if self.task_executor_max_thread >= self.executor.queue_size():
            sqs_messages = self.sqs_extended_client.receive_messages(
                self.sqs_primary_url,
                self.sqs_batch_size,
                self.visibility_time,
                self.sqs_wait_time,
            )
            self.logger.debug("%s Message Count [%s]"), thread_id, {len(sqs_messages)}
            if sqs_messages:
                for message in sqs_messages:
                    message_id = message["sqs"]["MessageId"]
                    self.logger.info("%s Message id [%s]", thread_id, message_id)
                    read_msg_ids.append(message["sqs"]["MessageId"])
                    task = self.executor.submit(self.message_executor, message)
                    running_threads.append(task)
        else:
            self.logger.info("%s Max thread limit reached hence stop reading messages from queue", thread_id)

        self.logger.debug(
            "%s Received Message count [%s] Created thread count [%s]",
            thread_id,
            len(read_msg_ids),
            len(running_threads),
        )
        return running_threads, read_msg_ids

    def _handle_message_delete_processing(self, running_threads, read_msg_ids):
        """
        Handling the message delete process from the SQS and S3 bucket if it is completed
        :param read_msg_ids All the message ids that taken from SQS
        :param running_threads Executing tasks to send emails
        """
        current_thread = threading.current_thread()
        thread_id = f"[{current_thread.name}:{current_thread.ident}]"
        receipt_handles_to_delete = []
        completed_msg_ids = []
        for future in as_completed(running_threads):
            try:
                msg = future.result()
                msg_id = msg["sqs"]["MessageId"]
                receipt_handles_to_delete.append(msg["sqs"])
                completed_msg_ids.append(msg_id)
                self.logger.debug("%s Execution completed and deleted from queue: %s", thread_id, msg_id)
            except Exception as e:
                self.logger.error("%s An error occurred while processing the message %s", thread_id, str(e))
        dif_msg_ids = [i for i in read_msg_ids if i not in completed_msg_ids]
        self.logger.debug("No of messages not processed [%s] and msg ids are %s", len(dif_msg_ids), dif_msg_ids)
        if receipt_handles_to_delete:
            self.sqs_extended_client.delete_messages(self.sqs_primary_url, receipt_handles_to_delete)



import os
from unittest import mock
from unittest.mock import ANY, Mock

from fsd_utils.healthchecks.healthcheck import Healthcheck


class TestHealthcheck:
    @mock.patch.dict(os.environ, clear=True)
    def testHealthChecksSetup(self):
        test_app = Mock()
        health = Healthcheck(test_app)
        test_app.add_url_rule.assert_called_with("/healthcheck", view_func=ANY, host="<host>")
        assert health.checkers == [], "Checks not initialised"

    @mock.patch.dict(os.environ, clear=True)
    def testWithNoChecks(self):
        mock_app = Mock()
        health = Healthcheck(mock_app)
        mock_app.add_url_rule.assert_called_with("/healthcheck", view_func=ANY, host="<host>")

        expected_dict = {"checks": []}

        result = health.healthcheck_view()
        assert result[0] == expected_dict, "Unexpected response body"
        assert result[1] == 200, "Unexpected status code"

    @mock.patch.dict(os.environ, clear=True)
    def testWithChecksPassing_mocks(self, flask_test_client):
        test_app = Mock()
        health = Healthcheck(test_app)
        test_app.add_url_rule.assert_called_with("/healthcheck", view_func=ANY, host="<host>")

        expected_dict = {"checks": [{"check_a": "ok"}, {"check_b": "ok"}]}

        check_a = Mock()
        check_a.check.return_value = True, "ok"
        check_a.name = "check_a"
        health.add_check(check_a)
        check_b = Mock()
        check_b.check.return_value = True, "ok"
        check_b.name = "check_b"
        health.add_check(check_b)

        result = health.healthcheck_view()
        assert result[0] == expected_dict, "Unexpected response body"
        assert result[1] == 200, "Unexpected status code"

    @mock.patch.dict(os.environ, clear=True)
    def testWithChecksFailing_mocks(self, flask_test_client):
        test_app = Mock()
        health = Healthcheck(test_app)
        test_app.add_url_rule.assert_called_with("/healthcheck", view_func=ANY, host="<host>")

        expected_dict = {"checks": [{"check_a": "fail"}, {"check_b": "ok"}]}

        check_a = Mock()
        check_a.check.return_value = False, "fail"
        check_a.name = "check_a"
        health.add_check(check_a)
        check_b = Mock()
        check_b.check.return_value = True, "ok"
        check_b.name = "check_b"
        health.add_check(check_b)

        result = health.healthcheck_view()
        assert result[0] == expected_dict, "Unexpected response body"
        assert result[1] == 500, "Unexpected status code"

    @mock.patch.dict(os.environ, clear=True)
    def testWithChecksException_mocks(self, flask_test_client):
        test_app = Mock()
        health = Healthcheck(test_app)
        test_app.add_url_rule.assert_called_with("/healthcheck", view_func=ANY, host="<host>")

        expected_dict = {"checks": [{"check_a": "fail"}, {"check_b": "Failed - check logs"}]}

        check_a = Mock()
        check_a.check.return_value = False, "fail"
        check_a.name = "check_a"
        health.add_check(check_a)
        check_b = Mock()
        check_b.check.side_effect = TypeError
        check_b.name = "check_b"
        health.add_check(check_b)

        result = health.healthcheck_view()
        assert result[0] == expected_dict, "Unexpected response body"
        assert result[1] == 500, "Unexpected status code"

    @mock.patch.dict(os.environ, {"GITHUB_SHA": "test-version"})
    def testShowsVersion(self):
        test_app = Mock()
        health = Healthcheck(test_app)
        result = health.healthcheck_view()
        assert result[0]["version"] == "test-version"



"""
Contains test configuration.
"""

import unittest.mock as mock
import uuid
from pathlib import Path

import pytest
from flask import Flask, g

from fsd_utils.authentication.config import SupportedApp
from fsd_utils.authentication.decorators import check_internal_user, login_requested, login_required


def create_app():
    app = Flask("test")
    return app


@pytest.fixture()
def app_context():
    with create_app().app_context():
        yield


@pytest.fixture(scope="function")
def flask_test_client():
    """
    Creates the test client we will be using to test the responses
    from our app, this is a test fixture.
    :return: A flask test client.
    """

    with create_app().app_context() as app_context:
        _test_public_key_path = str(Path(__file__).parent) + "/keys/rsa256/public.pem"
        with open(_test_public_key_path, mode="rb") as public_key_file:
            rsa256_public_key = public_key_file.read()

        app_context.app.config.update(
            {
                "FSD_LANG_COOKIE_NAME": "language",
                "COOKIE_DOMAIN": None,
                "FSD_USER_TOKEN_COOKIE_NAME": "fsd-user-token",
                "AUTHENTICATOR_HOST": "https://authenticator",
                "RSA256_PUBLIC_KEY": rsa256_public_key,
            }
        )
        app_context.app.add_url_rule(
            "/mock_login_required_route",
            "mock_login_required_route",
            mock_login_required_route,
        )
        app_context.app.add_url_rule(
            "/mock_login_requested_route",
            "mock_login_requested_route",
            mock_login_requested_route,
        )
        app_context.app.add_url_rule(
            "/mock_login_required_roles_route",
            "mock_login_required_roles_route",
            mock_login_required_roles_route,
        )
        app_context.app.add_url_rule(
            "/mock_login_required_admin_roles_route",
            "mock_login_required_admin_roles_route",
            mock_login_required_admin_roles_route,
        )

        app_context.app.add_url_rule(
            "/mock_login_requested_return_app_route",
            "mock_login_requested_return_app_route",
            mock_login_requested_return_app_route,
        )

        with app_context.app.test_client() as test_client:
            yield test_client


@pytest.fixture(scope="function")
def flask_test_development_client():
    """
    Creates the test client we will be using to test the responses
    from our app, this is a test fixture.
    :return: A flask test client.
    """

    with create_app().app_context() as app_context:
        _test_public_key_path = str(Path(__file__).parent) + "/keys/rsa256/public.pem"
        with open(_test_public_key_path, mode="rb") as public_key_file:
            rsa256_public_key = public_key_file.read()

        app_context.app.config.update(
            {
                "FSD_LANG_COOKIE_NAME": "language",
                "COOKIE_DOMAIN": None,
                "FLASK_ENV": "development",
                "DEBUG_USER_ON": True,
                "DEBUG_USER_ROLE": "ADMIN",
                "DEBUG_USER": {
                    "full_name": "Development User",
                    "email": "dev@example.com",
                    "roles": ["COF_ADMIN", "COF_TEST"],
                    "highest_role_map": {"COF": "ADMIN"},
                },
                "FSD_USER_TOKEN_COOKIE_NAME": "fsd-user-token",
                "AUTHENTICATOR_HOST": "https://authenticator",
                "RSA256_PUBLIC_KEY": rsa256_public_key,
            }
        )
        app_context.app.add_url_rule(
            "/mock_login_required_roles_route",
            "mock_login_required_roles_route",
            mock_login_required_roles_route,
        )
        app_context.app.add_url_rule(
            "/mock_login_required_admin_roles_route",
            "mock_login_required_admin_roles_route",
            mock_login_required_admin_roles_route,
        )
        with app_context.app.test_client() as test_client:
            yield test_client


@pytest.fixture(scope="function")
def flask_test_check_internal_user_client():
    """
    Creates the test client for check_internal_user tests.
    """
    with create_app().app_context() as app_context:
        _test_public_key_path = str(Path(__file__).parent) + "/keys/rsa256/public.pem"
        with open(_test_public_key_path, mode="rb") as public_key_file:
            rsa256_public_key = public_key_file.read()

        app_context.app.config.update(
            {
                "FSD_LANG_COOKIE_NAME": "language",
                "COOKIE_DOMAIN": None,
                "FSD_USER_TOKEN_COOKIE_NAME": "fsd-user-token",
                "AUTHENTICATOR_HOST": "https://authenticator",
                "RSA256_PUBLIC_KEY": rsa256_public_key,
            }
        )

        app_context.app.add_url_rule(
            "/mock_check_internal_user_route",
            "mock_check_internal_user_route",
            mock_check_internal_user_route,
            methods=["GET"],
        )

        with app_context.app.test_client() as test_client:
            yield test_client


@login_required
def mock_login_required_route():
    """
    A mock route function decorated with @login_required
    Here we expect a non logged in user to be redirected
    to authenticator, and a logged in user to have the required
    Flask request g variables set as below:
    g: {
        "is_authenticated": True,
        "logout_url": "https://authenticator/sessions/sign-out",
        "account_id": "test-user",
        "user": User(
            email="test@example.com",
            full_name="Test User",
            highest_role_map={"COF": "LEAD_ASSESSOR"},
            roles=["COF_LEAD_ASSESSOR", "COF_ASSESSOR", "COF_COMMENTER"]
        )
    :return: the Flask g variable serialised as a dict/json
    """
    return vars(g)


@login_requested
def mock_login_requested_route():
    """
    A mock route function decorated with @login_requested
    Here we expect a logged in user to have the required
    Flask request g variables set as below:
    g: {
        "is_authenticated": True,
        "logout_url": "https://authenticator/sessions/sign-out",
        "account_id": "test-user",
        "user": User(
            email="test@example.com",
            full_name="Test User",
            highest_role_map={"COF": "LEAD_ASSESSOR"},
            roles=["COF_LEAD_ASSESSOR", "COF_ASSESSOR", "COF_COMMENTER"]
        )
    and a non logged in user to have the Flask request g variables
    set as below:
    g: {
        "is_authenticated": False,
        "logout_url": "https://authenticator/sessions/sign-out",
        "account_id": None,
    }
    :return: the Flask g variable serialised as a dict/json
    """
    return vars(g)


@login_required(roles_required=["COF_COMMENTER"])
def mock_login_required_roles_route():
    """
    A mock route function decorated with
    @login_required(roles_required=["COMMENTER"])
    Here we expect a logged in user without the "COMMENTER"
    role to be redirected to a missing roles required
    error page on authenticator,
    and a logged in user WITH the "COMMENTER" role
    to have the required
    Flask request g variables set as below:
    g: {
        "is_authenticated": True,
        "logout_url": "https://authenticator/sessions/sign-out",
        "account_id": "test-user",
        "user": User(
            email="test@example.com",
            full_name="Test User",
            highest_role_map={"COF": "LEAD_ASSESSOR"},
            roles=["COF_LEAD_ASSESSOR", "COF_ASSESSOR", "COF_COMMENTER"]
        )
    :return: the Flask g variable serialised as a dict/json
    """
    return vars(g)


@login_required(roles_required=["COF_ADMIN", "COF_TEST"])
def mock_login_required_admin_roles_route():
    """
    A mock route function decorated with
    @login_required(roles_required=["ADMIN","TEST"])
    Here we expect a logged in user without the "ADMIN","TEST"
    role to be redirected to a missing roles required
    error page on authenticator,
    and a logged in user WITH BOTH the "ADMIN" and "TEST" roles
    to have the required
    Flask request g variables set as below:
    g: {
        "is_authenticated": True,
        "logout_url": "https://authenticator/sessions/sign-out",
        "account_id": "test-user",
        "user": User(
            email="test@example.com",
            full_name="Test User",
            highest_role_map={"COF": "LEAD_ASSESSOR"},
            roles=["COF_LEAD_ASSESSOR", "COF_ASSESSOR", "COF_COMMENTER"]
        )
    :return: the Flask g variable serialised as a dict/json
    """
    return vars(g)


@login_required(return_app=SupportedApp.POST_AWARD_FRONTEND)
def mock_login_requested_return_app_route():
    """
    A mock route function decorated with @login_requested
    Here we expect a logged in user to have the required
    Flask request g variables set as below:
    NOTE: the logout_url should end with the query string "?return_app=post-award-frontend" in both cases.
    g: {
        "is_authenticated": True,
        "logout_url": "https://authenticator/sessions/sign-out?return_app=post-award-frontend",
        "account_id": "test-user",
        "user": User(
            email="test@example.com",
            full_name="Test User",
            highest_role_map={"COF": "LEAD_ASSESSOR"},
            roles=["COF_LEAD_ASSESSOR", "COF_ASSESSOR", "COF_COMMENTER"]
        )
    and a non logged in user to have the Flask request g variables
    set as below:
    g: {
        "is_authenticated": False,
        "logout_url": "https://authenticator/sessions/sign-out?return_app=post-award-frontend",
        "account_id": None,
    }
    :return: the Flask g variable serialised as a dict/json
    """
    return vars(g)


@login_requested
@check_internal_user
def mock_check_internal_user_route():
    return {"status": "ok"}


@pytest.fixture(scope="function")
def mock_uuid4():
    with mock.patch("uuid.uuid4", return_value=uuid.UUID("6ba7b810-9dad-11d1-80b4-00c04fd430c8")):
        yield



from babel import negotiate_locale
from flask import request

from fsd_utils import CommonConfig


def get_lang():
    # get lang if lang query arg is set
    language_from_query_args = request.args.get("lang")
    if language_from_query_args:
        if language_from_query_args not in ["cy", "en"]:
            return "en"
        return language_from_query_args

    # get locale from cookie if set
    locale_from_cookie = request.cookies.get(CommonConfig.FSD_LANG_COOKIE_NAME)
    if locale_from_cookie:
        if locale_from_cookie not in ["cy", "en"]:
            return "en"
        return locale_from_cookie

    # otherwise guess preference based on user accept header
    preferred = [accept_language.replace("-", "_") for accept_language in request.accept_languages.values()]
    negotiated_locale = negotiate_locale(preferred, ["en", "cy"])
    if negotiated_locale:
        return negotiated_locale

    # default is to return english
    return "en"



from fsd_utils.authentication.utils import get_highest_role_map


def test_get_highest_role_map():
    roles = [
        "COF_ASSESSOR",  # highest role is assessor
        "COF_COMMENTER",
        "COF_SCOTLAND",
        "COF_ENGLAND",
        "COF_WALES",
        "COF_NORTHERNIRELAND",
        "NSTF_LEAD_ASSESSOR",  # highest role is lead assessor
        "NSTF_ASSESSOR",
        "NSTF_COMMENTER",
        "MOCKFUND_COMMENTER",  # highest role is commenter
        "INVALIDFUND_INVALIDROLE",  # invalid role is ignored
    ]

    result = get_highest_role_map(roles)

    assert "INVALIDFUND" not in result
    assert result == {
        "COF": "ASSESSOR",
        "NSTF": "LEAD_ASSESSOR",
        "MOCKFUND": "COMMENTER",
    }



import json
from datetime import datetime

import boto3
from botocore.exceptions import ClientError


class SQSClient:
    def __init__(
        self,
        aws_access_key_id=None,
        aws_secret_access_key=None,
        region_name="us-west-1",
        endpoint_url=None,
        **kwargs,
    ):
        if aws_access_key_id and aws_secret_access_key:
            self.client = boto3.client(
                "sqs",
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key,
                region_name=region_name,
                endpoint_url=endpoint_url,
                **kwargs,
            )
        else:
            # if 'aws_access_key_id' and 'aws_access_key_id' are not provided make sure to provide
            # 'AWS_ACCESS_KEY_ID' and 'AWS_SECRET_ACCESS_KEY' with environment variables
            self.client = boto3.client(
                "sqs",
                region_name=region_name,
                endpoint_url=endpoint_url,
                **kwargs,
            )

    def get_queues(self, prefix=None):
        """
        Gets a list of SQS queues. When a prefix is specified, only queues with names
        that start with the prefix are returned.

        :param prefix: The prefix used to restrict the list of returned queues.
        :return: A list of Queue names.
        """
        if prefix:
            queues = self.client.list_queues(QueueNamePrefix=prefix)["QueueUrls"]
        else:
            queues = self.client.list_queues()["QueueUrls"]
        if queues:
            print(f"Got queues: {', '.join([q for q in queues])}")
            queue_names = [url.split("/")[-1] for url in queues]
            return queue_names
        else:
            print("No queues found.")
            return []

    def remove_queue(self, queue_url):
        """
        Removes an SQS queue. When run against an AWS account, it can take up to
        60 seconds before the queue is actually deleted.

        :param queue: The queue to delete.
        :return: None
        """
        try:
            self.client.delete_queue(QueueUrl=queue_url)
            print(f"Deleted queue with URL={queue_url}.")
        except ClientError as error:
            print(f"Couldn't delete queue with URL={queue_url}!")
            raise error

    def get_queue_url(self, queue_name):
        """Return queue url for the given queue name."""
        response = self.client.get_queue_url(
            QueueName=queue_name,
        )
        return response["QueueUrl"]

    def submit_single_message(
        self,
        queue_url,
        message,
        extra_attributes: dict = None,
        message_group_id=None,
        message_deduplication_id=None,
    ):
        try:
            SQS_CUSTOM_ATTRIBUTES = {
                "message_created_at": {
                    "StringValue": str(datetime.now()),
                    "DataType": "String",
                },
            }
            # add extra message attributes (if provided)
            if extra_attributes:
                for key, value in extra_attributes.items():
                    SQS_CUSTOM_ATTRIBUTES[key] = value

            print(f"Attempting to place message on queue '{queue_url}'.")

            response = self.client.send_message(
                QueueUrl=queue_url,
                MessageBody=json.dumps(message),
                MessageAttributes=SQS_CUSTOM_ATTRIBUTES,
                MessageGroupId=message_group_id,
                MessageDeduplicationId=message_deduplication_id,
            )
            message_id = response["MessageId"]
            print(f"Message (id: {message_id}) submitted to queue: {queue_url}.")
            return message_id
        except Exception as e:
            print(
                "Error whilst staging onto queue"
                f" '{queue_url}', message with"
                f" attributes '{str(extra_attributes)}'."
                f" Error : {str(e)}"
            )
            return str(e), 500, {"x-error": "Error"}

    def submit_message(self, queue_url, messages, DelaySeconds=1):
        """
        Send a batch of messages in a single request to an SQS queue.
        This request may return overall success even when some messages were not sent.
        The caller must inspect the Successful and Failed lists in the response and
        resend any failed messages.

        :param queue_url: SQS Queue url.
        :param queue: The queue to receive the messages.
        :param messages: The messages to send to the queue. These are simplified to
                        contain only the message body and attributes.
        :return: The response from SQS that contains the list of successful and failed
                messages.
        """
        try:
            entries = [
                {
                    "Id": str(ind),
                    "MessageBody": msg["body"],
                    "MessageAttributes": msg["attributes"],
                    "DelaySeconds": DelaySeconds,
                }
                for ind, msg in enumerate(messages)
            ]
            response = self.client.send_message_batch(
                QueueUrl=queue_url,
                Entries=entries,
            )
            if "Successful" in response:
                for msg_meta in response["Successful"]:
                    print(f"Message sent to the queue {queue_url}, MessageId: {msg_meta['MessageId']}")
            if "Failed" in response:
                for msg_meta in response["Failed"]:
                    print(
                        f"Failed to send messages to queue: {queue_url}, "
                        f"attributes {messages[int(msg_meta['Id'])]['attributes']}"
                    )
        except ClientError as error:
            print(f"Send messages failed to queue: {queue_url}")
            raise error
        else:
            return response

    def receive_messages(self, queue_url, max_number, visibility_time=1, wait_time=1):
        """
        Receive a batch of messages in a single request from an SQS queue.

        :param queue_url: SQS Queue url
        :param max_number: The maximum number of messages to receive. The actual number
                        of messages received might be less.
        :param visibility_time: The maximum time for message to temporarily invisible to other receivers.
                                This gives the initial receiver a chance to process the message. If the receiver
                                successfully processes and deletes the message within the visibility timeout,
                                the message is removed from the queue.
        :param wait_time: The maximum time to wait (in seconds) before returning. When
                        this number is greater than zero, long polling is used. This
                        can result in reduced costs and fewer false empty responses.
        :return: The list of Message objects received. These each contain the body
                of the message and metadata and custom attributes.
        """
        try:
            response = self.client.receive_message(
                QueueUrl=queue_url,
                AttributeNames=["SentTimestamp", "ApproximateReceiveCount"],
                MessageAttributeNames=["All"],
                MaxNumberOfMessages=max_number,
                VisibilityTimeout=visibility_time,
                WaitTimeSeconds=wait_time,
            )
            if "Messages" in response.keys():
                messages = response["Messages"]
            elif response["ResponseMetadata"]["HTTPStatusCode"] == 200:
                print(f"No more messages available in queue: {queue_url}")
                return []

            for msg in messages:
                print(f"Received message ID: {msg['MessageId']}, Attributes: {msg['MessageAttributes']}")
        except Exception as error:
            print(f"Couldn't receive messages from queue: {queue_url} Error: {error}")
            raise error
        else:
            return messages

    def delete_messages(self, queue_url, message_receipt_handles):
        """
        Delete a batch of messages from a queue in a single request.

        :param queue_url: SQS Queue url
        :param message_receipt_handles: The list of messages handles to delete.
        :return: The response from SQS that contains the list of successful and failed
                message deletions.
        """
        try:
            entries = [
                {"Id": str(ind), "ReceiptHandle": receipt_handle}
                for ind, receipt_handle in enumerate(message_receipt_handles)
            ]
            response = self.client.delete_message_batch(QueueUrl=queue_url, Entries=entries)

            if "Successful" in response:
                for msg_meta in response["Successful"]:
                    print(f"Deleted {message_receipt_handles[int(msg_meta['Id'])]}")
            if "Failed" in response:
                for msg_meta in response["Failed"]:
                    print(f"Could not delete {message_receipt_handles[int(msg_meta['Id'])]}")
        except ClientError:
            print(f"Couldn't delete message from queue {queue_url}")
        else:
            return response

    def create_sqs_queue(self, queue_name, has_dlq=False, dlq_queue_name=None, max_recieve_count=3):
        """
        Creates an Amazon SQS & DLQ queue.

        :param queue_name: SQS queue name to create
        :param has_dlq: If true, DLQ(Dead letter queue) is created for given queue
        :param dlq_queue_name: If `has_dlq` is true, then DLQ queue name to create
        :param max_recieve_count: If `has_dlq` is true, then provide the max recieve count
                                  before moving messages to DLQ
        :return: (sqs_queue_url) URL of the queue created.
        """
        # get queue list
        queue_list = self.get_queues()

        # create SQS queue if not exists
        if queue_name not in queue_list:
            sqs_queue_url = self.client.create_queue(
                QueueName=queue_name,
            )["QueueUrl"]
            print(f"Successfully created SQS queue '{queue_name}'")
        else:
            print(f"SQS queue '{queue_name}' already exists!")
            sqs_queue_url = self.client.get_queue_url(QueueName=queue_name)["QueueUrl"]

        if has_dlq:
            # create DLQ queue if not exists
            if dlq_queue_name not in queue_list:
                dlq_queue_url = self.client.create_queue(
                    QueueName=dlq_queue_name,
                )["QueueUrl"]
                dlq_queue_arn = self.client.get_queue_attributes(QueueUrl=dlq_queue_url, AttributeNames=["QueueArn"])[
                    "Attributes"
                ]["QueueArn"]
            else:
                print(f"DLQ '{dlq_queue_name}' already exists!")
                dlq_queue_url = self.client.get_queue_url(QueueName=dlq_queue_name)["QueueUrl"]
                dlq_queue_arn = self.client.get_queue_attributes(QueueUrl=dlq_queue_url, AttributeNames=["QueueArn"])[
                    "Attributes"
                ]["QueueArn"]

            redrive_policy = {
                "deadLetterTargetArn": dlq_queue_arn,
                "maxReceiveCount": max_recieve_count,
            }

            self.client.set_queue_attributes(
                QueueUrl=sqs_queue_url,
                Attributes={"RedrivePolicy": json.dumps(redrive_policy)},
            )

        return sqs_queue_url

    def set_queue_attributes(self, queue_name: str = None, queue_url: str = None, attributes: dict = None):
        if not queue_url:
            queue_url = self.client.get_queue_url(QueueName=queue_name)["QueueUrl"]

        if not attributes:
            attributes = {}

        self.client.set_queue_attributes(
            QueueUrl=queue_url,
            Attributes=attributes,
        )


# Uncomment the below code to test the SQS client usage
# if __name__ == "__main__":

#     sqs_client = SQSClient(
#         aws_access_key_id="FSDIOSFODNN7EXAMPLE",  # pragma: allowlist secret
#         aws_secret_access_key="fsdlrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY",  # pragma: allowlist secret
#         region_name="eu-west-2",
#         endpoint_url="http://localhost:4566",
#     )

#     # display all queues
#     print(sqs_client.get_queues())

#     # create my_queue
#     my_queue_url = sqs_client.create_sqs_queue(queue_name="my_queue")
#     print(sqs_client.get_queues())

#     # get queue url
#     print(sqs_client.get_queue_url("my_queue"))

#     # delete my_queue
#     print(sqs_client.remove_queue(my_queue_url))

#     # create my_queue & my_queue_dlq
#     my_queue_url = sqs_client.create_sqs_queue(
#         queue_name="my_queue", has_dlq=True, dlq_queue_name="my_queue_dlq"
#     )
#     print(sqs_client.get_queues())

#     # delete my_queue & my_queue_dlq
#     print(sqs_client.remove_queue(my_queue_url))
#     print(sqs_client.remove_queue(sqs_client.get_queue_url("my_queue_dlq")))
#     print(sqs_client.get_queues())



from functools import wraps
from typing import List
from urllib.parse import urlencode

from flask import Request, abort, current_app, g, redirect, request
from jwt import ExpiredSignatureError, PyJWTError
from werkzeug.exceptions import HTTPException

from fsd_utils.authentication.utils import validate_token_rs256

from .config import (
    InternalDomain,
    SupportedApp,
    config_var_auth_host,
    config_var_logout_url_override,
    config_var_user_token_cookie_name,
    signout_route,
    user_route,
)
from .models import User


def _failed_redirect(return_app: SupportedApp | None):
    logout_url = _build_logout_url(return_app)
    return abort(redirect(logout_url))


def _failed_roles_redirect(roles_required: List[str]):
    authenticator_host = current_app.config[config_var_auth_host]

    params = {"roles_required": "|".join(roles_required)}

    return abort(redirect(authenticator_host + user_route + f"?{urlencode(params)}"))


def _check_access_token(return_app: SupportedApp | None = None, auto_redirect=True):
    """
    Check the expected auth cookie for a
    valid JWT token.
    :param auto_redirect: (bool) redirect on failed authentication if True
    :return:
        - If a valid JWT token is found then return token claims
        - If no valid token is found:
            -- If auto_redirect is True then issue a _failed_redirect()
            -- If auto_redirect is False then return False
    """
    user_token_cookie_name = current_app.config[config_var_user_token_cookie_name]

    login_cookie = request.cookies.get(user_token_cookie_name)
    if not login_cookie:
        if auto_redirect:
            _failed_redirect(return_app)
        return False

    try:
        return validate_token_rs256(login_cookie)
    except (PyJWTError, ExpiredSignatureError):
        if auto_redirect:
            _failed_redirect(return_app)
        return False


def _build_return_path(request: Request):
    query_string = ("?" + request.query_string.decode()) if request.query_string else ""
    return request.path + query_string


def _build_logout_url(return_app: SupportedApp | None):
    if override := current_app.config.get(config_var_logout_url_override):
        return override

    authenticator_host = current_app.config[config_var_auth_host]
    if return_app:
        return (
            authenticator_host
            + signout_route
            + f"?{urlencode({'return_app': return_app.value, 'return_path': _build_return_path(request)})}"
        )
    else:
        return authenticator_host + signout_route


def login_required(f=None, roles_required: List[str] = None, return_app: SupportedApp | None = None):
    """
     Execute function if request contains valid JWT
     and pass account auth params to route as attributes
     on the flask request global 'g' object:
         - g.account_id: (str) users account id
         - g.is_authenticated: (bool) authentication status
         - g.user - this holds a User object and associated attributes
         - g.logout_url: (str) the service sign-out url
     If no valid auth JWT found then redirect to
     service config invalid login route
    :param f:
    :param return_app: app to return to after login
    :param roles_required: (List(str), optional) a list of the
            roles required to access the decorated route
    :return:
    """
    if f is None:
        return lambda f: login_required(f=f, roles_required=roles_required, return_app=return_app)

    @wraps(f)
    def _wrapper(*args, **kwargs):
        try:
            token_payload = _check_access_token(return_app=return_app)
            g.account_id = token_payload.get("accountId")
            g.user = User.set_with_token(token_payload)
        except HTTPException as e:
            if current_app.config.get("FLASK_ENV") == "development" and current_app.config.get("DEBUG_USER_ON"):
                g.account_id = current_app.config.get("DEBUG_USER_ACCOUNT_ID")
                g.user = User(**current_app.config.get("DEBUG_USER"))
            else:
                raise e

        g.logout_url = _build_logout_url(return_app)
        g.is_authenticated = True
        if roles_required:
            if not any(role_required in g.user.roles for role_required in roles_required):
                _failed_roles_redirect(roles_required)
        return f(*args, **kwargs)

    return _wrapper


def login_requested(f):
    """
    If request contains valid JWT
    then pass account auth params to route as attributes
    on the flask request global 'g' object:
        - g.account_id: (str) users account id
        - g.is_authenticated: (bool) authentication status
        - g.user - this holds a User object and associated attributes
        - g.logout_url: (str) the service sign-out url
    If no valid auth JWT found then update the g.is_authenticated
    variable to False and the g.account_id to None
    """

    @wraps(f)
    def decorated(*args, **kwargs):
        token_payload = _check_access_token(auto_redirect=False)
        authenticator_host = current_app.config[config_var_auth_host]
        g.logout_url = authenticator_host + signout_route
        if token_payload and isinstance(token_payload, dict):
            g.account_id = token_payload.get("accountId")
            g.user = User.set_with_token(token_payload)
            g.is_authenticated = True
        else:
            g.account_id = None
            g.is_authenticated = False
        return f(*args, **kwargs)

    return decorated


def check_internal_user(func):
    @wraps(func)
    def decorated(*args, **kwargs):
        internal_domains = tuple(k.value for k in InternalDomain)
        authenticated = g.is_authenticated
        is_communities = g.is_authenticated and g.user.email.endswith(internal_domains)
        if not authenticated or is_communities:
            return func(*args, **kwargs)
        else:
            abort(403)

    return decorated



from __future__ import absolute_import

import datetime
import logging
import re
import time
from logging.config import dictConfig
from os import getpid
from threading import get_ident as get_thread_ident

from flask import Flask, current_app, request
from flask.ctx import has_request_context
from pythonjsonlogger.json import JsonFormatter as BaseJSONFormatter

# Log formats can use any attributes available in
# https://docs.python.org/3/library/logging.html#logrecord-attributes
LOG_FORMAT = "%(name)s %(levelname)s " "- %(message)s - from %(funcName)s in %(pathname)s:%(lineno)d"

_DEFAULT_FSD_LOG_LEVEL = "INFO"

DEV_DEBUG_LOG_FORMAT = "%(asctime)s %(levelname)s - %(message)s - from %(funcName)s() in" " %(filename)s:%(lineno)d"


# fields named in LOG_FORMAT and LOG_FORMAT_EXTRA_JSON_KEYS
# will always be included in json log output even if
# no such field was supplied in the log record,
# substituting a None value if necessary.
LOG_FORMAT_EXTRA_JSON_KEYS = ()


logger = logging.getLogger(__name__)


def _common_request_extra_log_context():
    return {
        "method": request.method,
        "url": request.url,
        "endpoint": request.endpoint,
        # pid and thread ident are both available on LogRecord by default,
        # as `process` and `thread` respectively, but I don't see a
        # straightforward way of selectively including them only in certain
        # log messages - they are designed to be included when the formatter
        # is being configured. This is why I'm manually grabbing them and
        # putting them in as `extra` here, avoiding the existing parameter
        # names to prevent LogRecord from complaining
        "process_": getpid(),
        # stringifying this as it could potentially be a long that json is
        # unable to represent accurately
        "thread_": str(get_thread_ident()),
    }


def get_default_logging_config(app: Flask):
    log_level = app.config.get("FSD_LOG_LEVEL", _DEFAULT_FSD_LOG_LEVEL)
    formatter = "plaintext" if app.config.get("FLASK_ENV") == "development" else "json"

    return {
        "version": 1,
        "disable_existing_loggers": False,
        "filters": {
            "request_extra_context": {
                "()": "fsd_utils.logging.logging.RequestExtraContextFilter",
            },
        },
        "formatters": {
            "plaintext": {
                "()": "fsd_utils.logging.logging.CustomLogFormatter",
                "msg": DEV_DEBUG_LOG_FORMAT,
            },
            "json": {
                "()": "fsd_utils.logging.logging.JSONFormatter",
                "fmt": get_json_log_format(),
            },
        },
        "handlers": {
            "null": {
                "class": "logging.NullHandler",
            },
            "default": {
                "filters": ["request_extra_context"],
                "formatter": formatter,
                "class": "logging.StreamHandler",
                "stream": "ext://sys.stdout",
            },
        },
        "loggers": {
            "": {
                "handlers": ["null"],
            },
            "werkzeug": {
                "disabled": True,
            },
            app.name: {
                "handlers": ["default"],
                "level": log_level,
            },
        },
    }


def init_app(app, log_config: dict | None = None):
    app.config.setdefault("FSD_LOG_LEVEL", _DEFAULT_FSD_LOG_LEVEL)
    log_config = log_config or get_default_logging_config(app)
    dictConfig(log_config)
    attach_request_loggers(app)
    app.logger.info("Logging configured")


def attach_request_loggers(app):
    @app.before_request
    def before_request():
        # annotating these onto request instead of flask.g as they probably
        # shouldn't be inheritable from a request-less application context
        request.before_request_real_time = time.perf_counter()
        request.before_request_process_time = time.process_time()

        current_app.logger.log(
            logging.DEBUG,
            "Received request {method} {url}",
            extra=_common_request_extra_log_context(),
        )

    @app.after_request
    def after_request(response):
        if request.path != "/healthcheck":
            current_app.logger.log(
                logging.INFO,
                "{method} {url} {status}",
                extra={
                    "status": response.status_code,
                    "duration_real": (
                        (time.perf_counter() - request.before_request_real_time)
                        if hasattr(request, "before_request_real_time")
                        else None
                    ),
                    "duration_process": (
                        (time.process_time() - request.before_request_process_time)
                        if hasattr(request, "before_request_process_time")
                        else None
                    ),
                    **_common_request_extra_log_context(),
                },
            )
        return response


def get_json_log_format():
    return LOG_FORMAT + "".join(f" %({key})s" for key in LOG_FORMAT_EXTRA_JSON_KEYS)


class BaseExtraStackLocationFilter(logging.Filter):
    def __init__(self, param_prefix):
        self._param_prefix = param_prefix

    def is_interesting_frame(self, frame):
        raise NotImplementedError

    def enabled_for_record(self, record):
        return True

    def _findCaller(self):
        """
        Cut down copy of Python 3.6.6's logging.Logger.findCaller()

        Find the stack frame of the caller so that we can note the source
        file name, line number and function name.
        """
        f = logging.currentframe()
        if f is not None:
            f = f.f_back
        while hasattr(f, "f_code"):
            co = f.f_code
            try:
                # do this in a try block to protect ourselves from
                # faulty is_interesting_frame implementations
                is_interesting_frame = self.is_interesting_frame(f)
            except:  # noqa
                is_interesting_frame = False
            if is_interesting_frame:
                return co.co_filename, f.f_lineno, co.co_name
            f = f.f_back
        return None, None, None

    def filter(self, record):
        if not self.enabled_for_record(record):
            return record

        rv = self._findCaller()

        if rv != (
            None,
            None,
            None,
        ):
            for attr_name, attr_value in zip(
                (
                    "pathname",
                    "lineno",
                    "funcName",
                ),
                rv,
                strict=False,
            ):
                setattr(record, self._param_prefix + attr_name, attr_value)

        return record


class RequestExtraContextFilter(logging.Filter):
    """
    Filter which will pull extra context from
    the current request's `get_extra_log_context` method
    (if present) and make this available on log records
    """

    def filter(self, record):
        if has_request_context() and callable(getattr(request, "get_extra_log_context", None)):
            for key, value in request.get_extra_log_context().items():
                setattr(record, key, value)

        return record


class CustomLogFormatter(logging.Formatter):
    """
    Accepts a format string for the message
    and formats it with the extra fields
    with console colouring option
    """

    FORMAT_STRING_FIELDS_PATTERN = re.compile(r"\((.+?)\)")
    GREY = "\x1b[38;20m"
    YELLOW = "\x1b[33;20m"
    RED = "\x1b[31;20m😠"
    BOLD_RED = "\x1b[31;1m"
    RESET = "\x1b[0m"

    COLOURS = {
        "DEBUG": GREY,
        "INFO": GREY,
        "WARNING": YELLOW,
        "ERROR": RED,
        "CRITICAL": BOLD_RED,
    }

    def __init__(self, msg, use_color=True):
        date_strftime_format = "%d-%b-%y %H:%M:%S"
        logging.Formatter.__init__(self, msg, datefmt=date_strftime_format)
        self.use_color = use_color

    def add_fields(self, record):
        """
        Ensure all values found in our `fmt`
        have non-None entries in `record`
        """
        # FIXME: This method mutates the log record's `levelname` attribute, which `colour_field` does a lookup on.
        #        This means if the log record is processed by two handlers, we get an error, because the second
        #        handler doesn't recognise the ANSI-encoded levelname. For now, we only process logs through a single
        #        handler - if this changes, beware, we'll be in uncharted (broken) territory.
        for field in self.FORMAT_STRING_FIELDS_PATTERN.findall(self._fmt):
            # slightly clunky - this is so we catch
            # explicitly-set Nones too and turn them into "-"
            fetched_value = record.__dict__.get(field)
            string_value = fetched_value if fetched_value is not None else "-"
            if self.use_color and field == "levelname":
                string_value = self.colour_field(record, string_value)
            record.__dict__[field] = string_value

    def colour_field(self, record, msg):
        color = self.COLOURS.get(record.levelname)
        return color + msg + self.RESET

    def format(self, record):
        self.add_fields(record)
        msg = super(CustomLogFormatter, self).format(record)

        try:
            msg = msg.format(**record.__dict__)
        except:  # noqa
            # We know that KeyError, ValueError and IndexError are all
            # possible things that can go wrong here - there is no guarantee
            # that the message passed into the logger is actually suitable
            # to be used as a format string. This is particularly so where
            # we are logging arbitrary exception that may reference code.
            #
            # We catch all exceptions rather than just those three, because
            # _any_ failure to format the message must not result in an error,
            # otherwise the original log message will never be returned and
            # written to the logs, and that might be important info such as an
            # exception.
            #
            # NB do not attempt to log either the exception or `msg` here,
            # or you will find that too fails, and you end up with an
            # infinite recursion / stack overflow.
            logger.info("failed to format log message")

        return msg


class JSONFormatter(BaseJSONFormatter):
    def __init__(self, *args, max_missing_key_attempts=5, **kwargs):
        super().__init__(*args, **kwargs)
        self._max_missing_key_attempts = max_missing_key_attempts

    def formatTime(self, record, datefmt: str | None = ...) -> str:
        ct = self.converter(record.created)
        if datefmt:
            s = time.strftime(datefmt, ct)
        else:
            s = (
                datetime.datetime.fromtimestamp(record.created, datetime.timezone.utc)
                .astimezone()
                .isoformat(sep=" ", timespec="milliseconds")
            )
        return s

    def process_log_record(self, log_record):
        for key, newkey in (
            (
                "asctime",
                "time",
            ),
            (
                "trace_id",
                "requestId",
            ),
        ):
            try:
                log_record[newkey] = log_record.pop(key)
            except KeyError:
                pass

        log_record["logType"] = "application"

        missing_keys = {}
        for _ in range(self._max_missing_key_attempts):
            try:
                log_record["message"] = log_record["message"].format(**log_record, **missing_keys)
            except KeyError as e:
                missing_keys[e.args[0]] = f"{{{e.args[0]}: missing key}}"
            else:
                # execution should only ever reach this point once
                # - when the .format() succeeds
                if missing_keys:
                    logger.warning("Missing keys when formatting log message: %s", tuple(missing_keys.keys()))

                break

        else:
            logger.exception("Too many missing keys when attempting to format log message:" " gave up")

        return log_record



import uuid

from fsd_utils.mapping.application.application_utils import number_to_month


class ProcessTypes:
    @classmethod
    def _str_single_item(cls, item: dict, sorted_data: dict) -> dict:
        """
        Convert a single item to a dictionary and append it to the sorted_data.

        Args:
            item (dict): A dictionary containing data.
            sorted_data (dict): A dictionary to store sorted data.

        Returns:
            dict: The updated sorted_data dictionary.
        """
        if len(item) < 2:
            for value in item.values():
                key = str(uuid.uuid4())
                sorted_data[key] = value
        if sorted_data:
            return sorted_data

    @classmethod
    def _dict_single_item(cls, item: dict, sorted_data: dict) -> dict:
        """
        Recursively processes a dictionary item to extract and validate date values.

        Args:
            item (dict): The dictionary item to process.
            sorted_data (dict): A dictionary to store sorted date values.

        Returns:
            dict: A dictionary containing validated date values.
        """
        for value in item.values():
            if isinstance(value, dict):
                date = cls.validated_iso_value(value)
                sorted_data[f"date_{str(uuid.uuid4())}"] = date
            elif isinstance(value, str):
                cls._str_single_item(item, sorted_data)
        if sorted_data:
            return sorted_data

    @classmethod
    def _dict_items(cls, item: dict, sorted_data: dict) -> dict:
        """
        Adds validated 'address' and 'value_str_len_two' items to a dictionary.

        This class method validates an 'item's address and checks if it has a two-item value string.
        If the address is valid, it is added to 'sorted_data' with a unique key.
        If 'value_str_len_two' is present, it is also added to 'sorted_data' with an appropriate key.
        If neither address nor 'value_str_len_two' are found, it sorts the dictionary items based on 'item'.

        Args:
            cls: The class reference.
            item: The item to be processed.
            sorted_data: The dictionary to which items are added.
        """
        address = cls.validate_address(item)
        if address:
            sorted_data[f"address_{str(uuid.uuid4())}"] = address

        value_len_two = cls.value_len_two(item)
        if value_len_two:
            sorted_data[value_len_two[0]] = value_len_two[1]

        if not address and not value_len_two:
            cls.sort_dict_items(item, sorted_data)

    @classmethod
    def sort_dict_items(cls, item: dict, sorted_data: dict) -> dict:
        """
        Sorts and updates a dictionary with a given item.

        Args:
            item: The item to be sorted and added to the dictionary.
            sorted_data: The dictionary to which the item will be added.

        Returns:
            dict: The updated dictionary with the sorted item.
        """

        _values, key = cls.value_len_greater_two(item)
        if len(_values) > 1:
            sorted_data[key] = _values
        else:
            sorted_data[key] = ", ".join(map(str, _values))

        if sorted_data:
            return sorted_data

    @classmethod
    def value_len_two(cls, item: dict) -> list:
        """
        Filters and returns a list of string values from a dictionary `item`
        if the dictionary has exactly two items and both values are strings.

        Parameters:
        - item (dict): The input dictionary to filter.

        Returns:
        - list: A list containing string values from the dictionary, or None if no such values are found.
        """
        _values = []
        if len(item) == 2:
            for value in item.values():
                if isinstance(value, (str, int)):
                    _values.append(value)
                if isinstance(value, dict):
                    date = cls.validated_iso_value(value)
                    if date:
                        _values.append(date)
                    address = cls.validate_address(value)
                    if address:
                        _values.append(address)
        if _values:
            return _values

    @classmethod
    def value_len_greater_two(cls, item: dict) -> tuple:
        """
        Extracts and combines values from a dictionary 'item' if its length is greater than 2.

        If 'item' contains dictionaries, it validates and appends ISO date values.
        If not, it combines string and integer values into a list along with a key.

        Args:
            item (dict): The dictionary to process.

        Returns:
            tuple: A tuple containing a list of combined values and the identified key, if available.
        """
        combined_values = []
        key = None
        if len(item) > 2:
            for value in item.values():
                if isinstance(value, dict):
                    date = cls.validated_iso_value(value)
                    if date:
                        combined_values.append(date)
                    address = cls.validate_address(value)
                    if address:
                        combined_values.append(address)
                    elif not date and not address:
                        dict_values = [val for val in value.values() if val is not None and val != ""]
                        combined_values.append(", ".join(map(str, dict_values)))
                if isinstance(value, (str, int)):
                    _str_key = [val for val in item.values() if isinstance(val, (str, int))]
                    if _str_key:
                        key = next(
                            (item for item in _str_key if not isinstance(item, int) and item is not None),
                            _str_key[0],
                        )

        _values = [val for val in item.values() if val != key if not isinstance(val, dict)]
        combined_values.extend(_values)

        if combined_values and key:
            return combined_values, key

    @classmethod
    def validated_iso_value(cls, value: dict) -> str:
        """
        Validates and converts ISO date components in 'value' dictionary to a formatted string.

        Args:
            value (dict): A dictionary containing ISO date components.

        Returns:
            str: A formatted string with valid date components, or None if no valid components found.
        """
        iso_keys = ["date", "month", "year"]
        date = []
        for k, v in value.items():
            key_split = k.split("__")
            if key_split:
                iso_keys_found = [n_key for n_key in key_split if n_key in iso_keys]
                if iso_keys_found:
                    month_name = number_to_month(
                        v,
                        *iso_keys_found,
                    )
                    date.append(month_name)
        if date:
            return " ".join(str(item) for item in date if item is not None)

    @classmethod
    def validate_address(cls, item: dict) -> list:
        """
        Validates and extracts a complete address from a dictionary.

        This class method takes a dictionary 'item' and checks for specific address keys
        ("addressLine1", "addressLine2", "county", "postcode", "town") to extract a
        complete address. If valid address components are found, they are returned as a list.

        Args:
            item (dict): A dictionary containing address components.

        Returns:
            list: A list of valid address components in the order of ["addressLine1",
                  "addressLine2", "county", "postcode", "town"]. Returns an empty list if
                  no valid components are found.
        """
        address_keys = ["addressLine1", "addressLine2", "county", "postcode", "town"]
        address = []
        for key, value in item.items():
            if key in address_keys and value is not None and value != "":
                address.append(value)
        if address:
            return ", ".join(str(item) for item in address if item is not None)

    @classmethod
    def is_valid_uuid(cls, uuid_str: str) -> bool:
        """
        Check if a given string contains valid UUIDs separated by underscores.

        Args:
            uuid_str (str): The input string possibly containing UUIDs.

        Returns:
            bool: True if at least one valid UUID is found in the string, otherwise False.
        """

        def is_valid_uuid_part(part):
            try:
                uuid.UUID(part, version=4)
                return True
            except ValueError:
                return False

        split_uuid_str = uuid_str.split("_")
        if any(part for part in split_uuid_str if is_valid_uuid_part(part)):
            return True

        return False



EN, CY = "en", "cy"
SUPPORTED_LANGUAGES = {EN, CY}

NOT_PROVIDED = {
    EN: "Not provided",
    CY: "Heb ei ddarparu",
}

YES = {
    EN: "Yes",
    CY: "Oes",
}

NO = {
    EN: "No",
    CY: "Nac Oes",
}



import pytest
from flask_migrate import upgrade
from sqlalchemy import text
from sqlalchemy_utils.functions import create_database, database_exists, drop_database


def prep_db(reuse_db=False):
    """
    Determines whether the target database already exists, if it
    doesn't it is created.
    If it exists but reuse_db is False, then the DB is dropped and recreated.
    upgrade() is always run to make sure the schema is up to date.
    """

    from config import Config

    no_db = not database_exists(Config.SQLALCHEMY_DATABASE_URI)
    refresh_db = not reuse_db

    if no_db:
        create_database(Config.SQLALCHEMY_DATABASE_URI)

    elif refresh_db:
        drop_database(Config.SQLALCHEMY_DATABASE_URI)
        create_database(Config.SQLALCHEMY_DATABASE_URI)

    upgrade()


@pytest.fixture(scope="session")
def _db(app, request):
    """
    Fixture to supply tests with direct access to the database
    """
    yield app.extensions["sqlalchemy"]


@pytest.fixture(scope="session")
def recreate_db(request, _db, app):
    """
    Fixture to determine whether we need to clear out the existing DB
    (based on cache values) and then call prep_db.
    This is session scoped so we only drop/create the db
    at the start of a test session
    """
    print("Recreating db")
    reuse_db = bool(request.config.cache.get("reuse_db", False))
    with app.app_context():
        prep_db(reuse_db)
    request.config.cache.set("reuse_db", True)
    yield


@pytest.fixture(scope="session")
def clear_test_data(app, _db, request, recreate_db):
    """
    Fixture to clean up the database after each test.

    This fixture reads preserve_test_data from the cache
    (see enable_preserve_test_data below),
    and if we are not preserving the data it clears the database
    by deleting all data from all tables.

    This is module scoped so that each test file gets a fresh empty
    database.

    """
    with app.app_context():
        yield
        preserve_test_data = request.config.cache.get("preserve_test_data", None)
        if not preserve_test_data:
            # rollback incase of any errors during test session
            _db.session.rollback()
            # disable foreign key checks
            _db.session.execute(text("SET session_replication_role = replica"))
            # delete all data from tables
            for table in reversed(_db.metadata.sorted_tables):
                _db.session.execute(table.delete())
            # reset foreign key checks
            _db.session.execute(text("SET session_replication_role = DEFAULT"))
            _db.session.commit()
        else:
            # If test requests 'preserve test data' make sure
            # on the next run we clear out the DB completely.
            request.config.cache.set("reuse_db", False)
            request.config.cache.set("preserve_test_data", False)


@pytest.fixture(scope="function")
def enable_preserve_test_data(request):
    """
    Fixture to read the markers on a test and if preserve_test_data is
    set, it sets it in the cache.
    This can't be combined with clear_test_data due to conflicts on the fixture
    scope, so this function reads the test markers and clear_test_data
    reads the cache.
    """
    marker = request.node.get_closest_marker("preserve_test_data")
    if marker is not None:
        request.config.cache.set("preserve_test_data", True)



import json

import pytest
from flask import Flask

from fsd_utils.logging import logging as fsd_logging
from fsd_utils.logging.logging import get_default_logging_config


@pytest.mark.parametrize(
    "fsd_log_level, expected_num_logs",
    (
        ("DEBUG", 6),  # 5 from test, +1 from `fsd_logging.init_app`
        ("INFO", 5),  # 4 from test, +1 from `fsd_logging.init_app`
        ("WARNING", 3),
        ("ERROR", 2),
        ("CRITICAL", 1),
    ),
)
def test_default_logging_respect_fsd_log_level(fsd_log_level, expected_num_logs, capsys):
    """This test captures logging output on stdout to ensure that our standard logging config respects
    FSD_LOG_LEVEL correctly. But this means that the test will fail if pytest is called with the `-s` flag,
    which disables capturing.

    So hopefully if you're reading this because the test has failed, and you're using the `-s` flag, you can
    be reassured that's why it's failing.
    """
    app = Flask("test_app")
    app.config["FSD_LOG_LEVEL"] = fsd_log_level

    fsd_logging.init_app(app)

    with app.app_context():
        app.logger.debug("debug")
        app.logger.info("info")
        app.logger.warning("warning")
        app.logger.error("error")
        app.logger.critical("critical")

    assert len(capsys.readouterr().out.strip().splitlines()) == expected_num_logs


def test_default_logging_is_json(capsys):
    """This test captures logging output on stdout to ensure that our standard logging config respects
    FSD_LOG_LEVEL correctly. But this means that the test will fail if pytest is called with the `-s` flag,
    which disables capturing.

    So hopefully if you're reading this because the test has failed, and you're using the `-s` flag, you can
    be reassured that's why it's failing.
    """
    app = Flask("test_app")
    app.config["FSD_LOG_LEVEL"] = "DEBUG"

    fsd_logging.init_app(app)

    log_lines = capsys.readouterr().out.strip().splitlines()
    assert len(log_lines) == 1

    assert isinstance(json.loads(log_lines[0]), dict)


def test_logging_is_not_json_if_development_flask_env(capsys):
    """This test captures logging output on stdout to ensure that our standard logging config respects
    FSD_LOG_LEVEL correctly. But this means that the test will fail if pytest is called with the `-s` flag,
    which disables capturing.

    So hopefully if you're reading this because the test has failed, and you're using the `-s` flag, you can
    be reassured that's why it's failing.
    """
    app = Flask("test_app")
    app.config["FLASK_ENV"] = "development"
    app.config["FSD_LOG_LEVEL"] = "DEBUG"

    fsd_logging.init_app(app)

    log_lines = capsys.readouterr().out.strip().splitlines()
    assert len(log_lines) == 1

    with pytest.raises(ValueError):
        assert json.loads(log_lines[0])


def test_init_app_logging_can_take_custom_logging_config(capsys):
    """This test captures logging output on stdout to ensure that our standard logging config respects
    FSD_LOG_LEVEL correctly. But this means that the test will fail if pytest is called with the `-s` flag,
    which disables capturing.

    So hopefully if you're reading this because the test has failed, and you're using the `-s` flag, you can
    be reassured that's why it's failing.
    """
    app = Flask("test_app")
    app.config["FLASK_ENV"] = "development"
    app.config["FSD_LOG_LEVEL"] = "WARNING"

    log_config = get_default_logging_config(app)
    fsd_logging.init_app(app, log_config=log_config)

    with app.app_context():
        app.logger.warning("warning")
        app.logger.critical("critical")

    assert len(capsys.readouterr().out.strip().splitlines()) == 2

    log_config = get_default_logging_config(app)
    log_config["loggers"]["test_app"]["level"] = "CRITICAL"
    fsd_logging.init_app(app, log_config=log_config)

    with app.app_context():
        app.logger.warning("warning")
        app.logger.critical("critical")

    assert len(capsys.readouterr().out.strip().splitlines()) == 1



import os


def configclass(cls):
    # Checks base classes for _config_info_, a dict containing
    # all infomation about config. attrs, and
    # adds them to a list.
    config_infos = [base.__dict__["_config_info_"] for base in cls.__bases__ if "_config_info_" in base.__dict__]

    # We compile all of these _config_info_'s into a dictionary,
    #  with overwritten values.
    settled_bases_config_info = {k: v for config_dict in config_infos for k, v in config_dict.items()}

    # We extract the values from the class and overlay
    #  the above dict with its values.
    for k in cls.__dict__:
        if not (k.startswith("__") or k.endswith("__")):
            settled_bases_config_info[k] = cls.__qualname__

    # This is the dicitonary that will be used
    #  by classes which inherit this class
    cls._config_info_ = settled_bases_config_info

    @classmethod
    def pretty_print(self):
        if os.environ.get("FSD_UTILS_ENABLE_CONFIG_TABLE") == "false":
            return

        from rich.console import Console
        from rich.table import Table

        table = Table(title="Config Info", show_lines=True)

        table.add_column("Key", justify="right", style="cyan", no_wrap=True)

        table.add_column("From", justify="right", style="yellow", no_wrap=True)

        for k, v in self._config_info_.items():
            config_key = str(k)
            from_value = str(v)
            table.add_row(config_key, from_value)

        console = Console()
        console.print(table)

    cls.pretty_print = pretty_print

    return cls



class NotifyConstants:
    # ---- Notification service endpoint ----#

    NOTIFICATION_SERVICE_ENDPOINT = "/send"

    # ---- Required keys Fields (DO NOT CHANGE) ----#

    FIELD_TYPE = "type"  # for TEMPLATE_TYPE.
    FIELD_CONTENT = "content"  # for contents to be shared.
    FIELD_TO = "to"  # for contact_info (do not set in common config)
    FIELD_FULL_NAME = "full_name"  # for contact_name (do not set in common config)

    # ---- Template types constants (DO NOT CHANGE) ---#

    TEMPLATE_TYPE_MAGIC_LINK = "MAGIC_LINK"
    TEMPLATE_TYPE_APPLICATION = "APPLICATION_RECORD_OF_SUBMISSION"
    TEMPLATE_TYPE_INCOMPLETE_APPLICATION = "INCOMPLETE_APPLICATION_RECORDS"
    TEMPLATE_TYPE_REMINDER = "APPLICATION_DEADLINE_REMINDER"
    TEMPLATE_TYPE_EOI_PASS = "Full pass"
    TEMPLATE_TYPE_EOI_PASS_W_CAVEATS = "Pass with caveats"
    TEMPLATE_TYPE_ASSESSMENT_APPLICATION_ASSIGNED = "TEMPLATE_TYPE_ASSESSMENT_APPLICATION_ASSIGNED"
    TEMPLATE_TYPE_ASSESSMENT_APPLICATION_UNASSIGNED = "TEMPLATE_TYPE_ASSESSMENT_APPLICATION_UNASSIGNED"

    # ---- Magic_Link Fields for FIELD_CONTENT ----#

    MAGIC_LINK_CONTACT_HELP_EMAIL_FIELD = "contact_help_email"
    MAGIC_LINK_URL_FIELD = "magic_link_url"
    MAGIC_LINK_REQUEST_NEW_LINK_URL_FIELD = "request_new_link_url"
    MAGIC_LINK_FUND_NAME_FIELD = "fund_name"

    # ---- Application Fields for FIELD_CONTENT ----#

    APPLICATION_FIELD = "application"
    APPLICATION_FORMS_FIELD = "forms"
    APPLICATION_NAME_FIELD = "name"
    APPLICATION_QUESTIONS_FIELD = "questions"
    APPLICATION_CAVEATS = "caveats"

    # To be discussed with Gio/Jack/Adam

    # APPLICATION_SUBMISSION_DATE_FIELD = "date_submitted"
    # APPLICATION_FUND_NAME_FIELD = "project_name"
    # APPLICATION_ROUND_NAME_FIELD = "round_id"
    # APPLICATION_ID_FIELD = "id"



from concurrent.futures import ThreadPoolExecutor
from contextvars import copy_context


class ContextAwareExecutor:
    """This Executor copy current flask application context and then inherit the
    flask application context for each executor thread then those threads will
    have the ability to use flask resources with its own flask context."""

    def __init__(self, max_workers, thread_name_prefix, flask_app):
        """Initialize Threadpool executor and ContextAwareExecutor :max_workers
        number of workers for the thread pool :thread_name_prefix prefix of the
        thread pool name :flask_app original flask application context."""
        self.executor = ThreadPoolExecutor(max_workers=max_workers, thread_name_prefix=thread_name_prefix)
        self.flask_app = flask_app

    def queue_size(self):
        """Get queue size of the Thread pool."""
        return self.executor._work_queue.qsize()

    def submit(self, fn, *args, **kwargs):
        """Submit executor to the thread pool."""
        ctx = copy_context()
        future = self.executor.submit(ctx.run, self.wrap_function(fn), *args, **kwargs)
        return future

    def wrap_function(self, fn):
        """Wrap the function with copied application context."""

        def wrapped(*args, **kwargs):
            with self.flask_app.app_context():
                return fn(*args, **kwargs)

        return wrapped



from flask import Flask
from flask_redis import FlaskRedis
from flipper import FeatureFlagClient, RedisFeatureFlagStore

redis_store = FlaskRedis(config_prefix="TOGGLES")


def initialise_toggles_redis_store(flask_app: Flask):
    redis_store.init_app(flask_app)


def create_toggles_client():
    store = RedisFeatureFlagStore(redis_store, base_key="feature")
    client = FeatureFlagClient(store)
    return client


def load_toggles(feature_configuration: dict, client: FeatureFlagClient):
    for feature, toggle in feature_configuration.items():
        client.create(feature)
        if toggle:
            client.enable(feature)
